@@ -14,11 +14,11 @@
 
 
 from pytket import OpType, Circuit
-from pytket.passes import RebaseUFR  # type: ignore
+from pytket.passes import DecomposeBoxes  # type: ignore
 from pytket.backends import Backend
-from pytket.utils import QubitPauliOperator
+from pytket.utils import QubitPauliOperator, get_operator_expectation_value
 from typing import List, Tuple
-import copy
+from copy import copy
 from qermit import (
     MitEx,
     ObservableTracker,
@@ -29,10 +29,21 @@ from qermit import (
     TaskGraph,
 )
 from qermit.taskgraph import gen_compiled_MitRes
-from .cdr_post import cdr_calibration_task_gen, cdr_correction_task_gen, _PolyCDRCorrect
+from .cdr_post import (
+    cdr_calibration_task_gen,
+    cdr_correction_task_gen,
+    _PolyCDRCorrect,
+    cdr_quality_check_task_gen,
+)
 import numpy as np
 import random
 from enum import Enum
+import warnings
+from pytket.passes import auto_rebase_pass
+
+
+ufr_gateset = {OpType.CX, OpType.Rz, OpType.H}
+ufr_rebase = auto_rebase_pass(ufr_gateset)
 
 
 class LikelihoodFunction(Enum):
@@ -56,7 +67,7 @@ class LikelihoodFunction(Enum):
 def sample_weighted_clifford_angle(rz_angle: float, **kwargs) -> float:
     """
     Calculates a weights distribution over different possible Clifford gates from input gate.
-    Clifford gates prepared by taking S gate to the power of n in {0,8}.
+    Clifford gates prepared by taking S gate to the power of n in {0,4}.
     n value sampled from calculated weights distribution.
     Distribution calculation as in B1, page 6 arXiv:2005.10189.
 
@@ -72,16 +83,19 @@ def sample_weighted_clifford_angle(rz_angle: float, **kwargs) -> float:
 
     rz_angle = rz_angle % 2
     rz_angle_matrix = np.asarray(
-        [[np.exp(-0.5 * np.pi * rz_angle), 0], [0, np.exp(0.5 * np.pi * rz_angle)]]
+        [
+            [np.exp(-0.5 * np.pi * rz_angle * 1j), 0],
+            [0, np.exp(0.5 * np.pi * rz_angle * 1j)],
+        ]
     )
     weights = []
-    for n in range(8):
+    for n in range(4):
         sn_matrix = np.asarray(
-            [[np.exp(-0.125 * np.pi * n), 0], [0, np.exp(0.125 * np.pi * n)]]
+            [[np.exp(-0.25 * np.pi * n * 1j), 0], [0, np.exp(0.25 * np.pi * n * 1j)]]
         )
         d = np.linalg.norm(rz_angle_matrix - sn_matrix)
-        weights.append(np.exp((-(d ** 2)) * 4))
-    return 0.25 * random.choices(range(8), weights)[0]
+        weights.append(np.exp((-(d**2)) * 4))
+    return 0.5 * random.choices(range(4), weights)[0]
 
 
 def gen_state_circuits(
@@ -89,7 +103,7 @@ def gen_state_circuits(
 ) -> List[Circuit]:
     """
     For given circuit c, returns total_state_circuits number of circuits, where each circuit is
-    run through some MitEx object to provide characteriastion data for later correction.
+    run through some MitEx object to provide characterisation data for later correction.
 
     State circuit construction as in appendix B of arXiv:2005.10189.
     State circuits are generated via a Markov Chain Monte Carlo technique.
@@ -110,7 +124,7 @@ def gen_state_circuits(
 
     :param c: Circuit for producing state circuits from.
     :type c: Circuit
-    :param n_non_cliffords: Number of non-Clifford gates in resulting characteriastion state circuits
+    :param n_non_cliffords: Number of non-Clifford gates in resulting characterisation state circuits
     :type n_non_cliffords: int
     :param n_pairs: Pairs of Clifford, Non-Clifford gates in state circuit generated.
     :type n_pairs: int
@@ -124,14 +138,16 @@ def gen_state_circuits(
     # set seed if given
     if "seed" in kwargs:
         random.seed(kwargs.get("seed"))
+        np.random.seed(kwargs.get("seed"))
 
     # Work in CX, H, Rz basis for ease
-    RebaseUFR().apply(c)
+    DecomposeBoxes().apply(c)
+    ufr_rebase.apply(c)
     c.flatten_registers()
     all_coms = c.get_commands()
 
     #  angles that make Clifford gates for S^n
-    clifford_angles = set({0.25, 0.5, 0.75, 1.0, 1.25, 1.5, 1.75, 0})
+    clifford_angles = set({0, 0.5, 1.0, 1.5, 2, 2.5, 3, 3.5})
 
     if n_pairs > n_non_cliffords:
         raise ValueError(
@@ -146,78 +162,112 @@ def gen_state_circuits(
         if all_coms[i].op.type == OpType.Rz:
             if all_coms[i].op.params[0] not in clifford_angles:
                 rz_ops.add(i)
+    if len(rz_ops) == 0:
+        return [c] * total_state_circuits
 
-    if n_pairs > len(rz_ops):
-        raise ValueError(
-            "More pairs {} than total Rz operations in circuit {}. Number of pairs must be less than or equal to.".format(
-                n_pairs, len(rz_ops)
-            )
-        )
-
-    # max_non_cliffs gives the number of gates to leave as non-Clifford in circuit
-    # State circuits are generated by sampling pairs of gates where
-    # all gates were originally non-Clifford, but some have been changed to Clifford
-    # therefore, there must be another Clifford gates remaining to produce all pairs
-    # this doesn't throw an error as this threshold is hard to predict in advance
-    max_non_cliffs = min(len(rz_ops) - n_pairs, n_non_cliffords)
-    # non_cliffords are indices for gates to be left non Clifford
-
-    non_cliffords = set(random.sample(rz_ops, max_non_cliffs))
-    # rz_ops then only contains rz gates in c to be substitued for Clifford angles
-    rz_ops.difference_update(non_cliffords)
-    # Power of random Clifford gates to be substitued
-    cliffords = {num: random.randint(0, 8) for num in rz_ops}
-
-    # TODO: Update this to use new circuit modification techniques available in pytket
     state_circuits: List[Circuit] = []
+    if len(rz_ops) == 1:
+        # make special case where its not possible to swap pairs
+        # produce state circuits until limit reached
+        while len(state_circuits) < total_state_circuits:
+            new_circuit = Circuit(c.n_qubits, len(c.bits))
+            for i in range(len(all_coms)):
+                com = all_coms[i]
+                if com.op.type == OpType.Rz:
+                    new_circuit.add_barrier(com.qubits)
+                    angle = sample_weighted_clifford_angle(com.op.params[0])
+                    new_circuit.add_gate(com.op.type, [angle], com.qubits)
+                    new_circuit.add_barrier(com.qubits)
+                # Measure gate has special case, but can assume 1 qubit to 1 bit
+                elif com.op.type is OpType.Measure:
+                    new_circuit.Measure(com.qubits[0], com.bits[0])
+                # A special case for Barrier metaop
+                elif com.op.type == OpType.Barrier:
+                    new_circuit.add_barrier(com.args)
+                # CX or H gate, add as is
+                else:
+                    new_circuit.add_gate(com.op.type, com.qubits)
+
+            # all circuits accepted and run, some results later discarded if not accepted by Metropolis-Hastings rule
+            state_circuits.append(new_circuit)
+        return state_circuits
+
+    # reassign variables where appropriate to guarantee there is always one pair of non-Clifford Clifford
+    # that can be resampled
+    n_non_cliffords = min(n_non_cliffords, len(rz_ops) - 1)
+    n_cliffords = len(rz_ops) - n_non_cliffords
+    n_pairs = min(n_cliffords, n_non_cliffords, n_pairs)
+
+    # non_cliffords are indices for gates to be left non Clifford
+    non_cliffords = np.random.choice(list(rz_ops), n_non_cliffords, replace=False)
+
+    # rz_ops then only contains rz gates in c to be substituted for Clifford angles
+    rz_ops.difference_update(non_cliffords)
+    # Power of random Clifford gates to be substituted
+    cliffords = {num: random.randint(0, 4) for num in rz_ops}
+
     # keep on producing state circuits until limit reached
     while len(state_circuits) < total_state_circuits:
         # cliffords.keys() are integers for now Clifford gates
         # sample some set of these to be subbed for original non-Clifford angle
-        clifford_pair_elements = set(random.sample(cliffords.keys(), n_pairs))
+        clifford_pair_elements = random.sample(list(cliffords.keys()), n_pairs)
         # from remaining non-Clifford Rz gates, sample some to have random Clifford gate
-        non_clifford_pair_elements = set(random.sample(non_cliffords, n_pairs))
+        non_clifford_pair_elements = random.sample(list(non_cliffords), n_pairs)
 
         # create new Circuit from scratch
         new_circuit = Circuit(c.n_qubits, len(c.bits))
         for i in range(len(all_coms)):
             com = all_coms[i]
             if com.op.type == OpType.Rz:
+                new_circuit.add_barrier(com.qubits)
                 # 3 sets of gates int must be in
                 # in clifford_pair_elements means gate has been denominated as Clifford,
                 # but is in some sampled pair so add original angle
                 if i in clifford_pair_elements:
                     new_circuit.add_gate(com.op.type, com.op.params, com.qubits)
+                    new_circuit.add_barrier(com.qubits)
                 # in non_clifford_pair_elements mean gate was denominated to be left non-Clifford,
                 # but its value has been sampled in a pair to now be Clifford
                 # random angle is sampled and returned
                 elif i in non_clifford_pair_elements:
                     angle = sample_weighted_clifford_angle(com.op.params[0])
                     new_circuit.add_gate(com.op.type, [angle], com.qubits)
+                    new_circuit.add_barrier(com.qubits)
                 # in cliffords mean it is denominated as Clifford, and hasn't been sampled for a pair
                 # as clifford_pair_elements has already been checked
                 # in this case, cliffords is a dict between Rz index and substitution S power
-                # get power from dict, multiply by 0.25 to get angle, add to ceircuit
+                # get power from dict, multiply by 0.5 to get angle, add to circuit
                 elif i in cliffords:
-                    new_circuit.add_gate(com.op.type, [0.25 * cliffords[i]], com.qubits)
+                    new_circuit.add_gate(com.op.type, [0.5 * cliffords[i]], com.qubits)
+                    new_circuit.add_barrier(com.qubits)
                 # final case means gate was chosen to retain non-Clifford, and has not been
                 # sampled in any pair, so add original angle.
                 else:
                     new_circuit.add_gate(com.op.type, com.op.params, com.qubits)
+                    new_circuit.add_barrier(com.qubits)
             # Measure gate has special case, but can assume 1 qubit to 1 bit
             elif com.op.type is OpType.Measure:
                 new_circuit.Measure(com.qubits[0], com.bits[0])
+            # A special case for Barrier metaop
+            elif com.op.type == OpType.Barrier:
+                new_circuit.add_barrier(com.args)
             # CX or H gate, add as is
             else:
                 new_circuit.add_gate(com.op.type, com.qubits)
 
         # all circuits accepted and run, some results later discarded if not accepted by Metropolis-Hastings rule
         state_circuits.append(new_circuit)
+
     return state_circuits
 
 
 def ccl_state_task_gen(
-    n_non_cliffords: int, n_pairs: int, total_state_circuits: int
+    n_non_cliffords: int,
+    n_pairs: int,
+    total_state_circuits: int,
+    simulator_backend: Backend,
+    tolerance: float,
+    max_state_circuits_attempts: int,
 ) -> MitTask:
     """
     Returns a MitTask object for which given some set of experiments,
@@ -231,6 +281,18 @@ def ccl_state_task_gen(
     :type n_pairs:
     :param total_state_circuits: Number of state circuits prepared for characterisation.
     :type total_state_circuits: int
+    :param tolerance: Model can be perturbed when calibration circuits have by
+        exact expectation values too close to each other. This parameter
+        sets a distance between exact expectation values which at least some
+        calibration circuits should have.
+    :type tolerance: float
+    :param simulator_backend: Backend object simulated characterisation experiments are
+        default run through.
+    :type simulator_backend: Backend
+    :param max_state_circuits_attempts: The maximum number of times to attempt to generate a
+        list of calibrations circuit with significantly different expectation
+        values, before resorting to a list with similar expectation values.
+    :type max_state_circuits_attempts: int
 
     :return: MitTask object for preparing and returning state circuits for characterisation.
     :rtype: MitTask
@@ -263,12 +325,36 @@ def ccl_state_task_gen(
             # generate all state circuits
             c_copy = ansatz_circuit.Circuit.copy()
             c_copy.symbol_substitution(ansatz_circuit.SymbolsDict._symbolic_map)
-            state_circuits = gen_state_circuits(
-                c_copy,
-                n_non_cliffords,
-                n_pairs,
-                total_state_circuits,
-            )
+
+            all_close = True
+            attempt = 0
+            while all_close and attempt < max_state_circuits_attempts:
+
+                state_circuits = gen_state_circuits(
+                    c_copy,
+                    n_non_cliffords,
+                    n_pairs,
+                    total_state_circuits,
+                )
+
+                pauli_expectation_list = [
+                    get_operator_expectation_value(
+                        c, qubit_pauli_operator, simulator_backend
+                    )
+                    for c in state_circuits
+                ]
+                all_close = all(
+                    abs(pauli_expectation - pauli_expectation_list[0]) <= tolerance
+                    for pauli_expectation in pauli_expectation_list
+                )
+
+                attempt += 1
+
+            if all_close:
+                warnings.warn(
+                    "Clifford Data Regression performs best when the exact expectation values of all calibration circuits are not the same. However, the generated calibration circuits have similar exact expectation values. Fit of the extrapolation function may be poor as a result."
+                )
+
             # for each state circuit, create a new wire of for each state circuit
             # one for simulator, one for device
             for c in state_circuits:
@@ -277,25 +363,23 @@ def ccl_state_task_gen(
                         Circuit=c, Shots=shots, SymbolsDict=SymbolsDict()
                     ),
                     ObservableTracker=ObservableTracker(
-                        copy.copy(qubit_pauli_operator)
-                    ),
+                        copy(qubit_pauli_operator)
+                    ),  # no copy means changes to one QubitPauliOperator can be made to all
                 )
                 wire_device = ObservableExperiment(
                     AnsatzCircuit=AnsatzCircuit(
                         Circuit=c.copy(),
-                        Shots=copy.copy(shots),
+                        Shots=copy(shots),
                         SymbolsDict=SymbolsDict(),
                     ),
-                    ObservableTracker=ObservableTracker(
-                        copy.copy(qubit_pauli_operator)
-                    ),
+                    ObservableTracker=ObservableTracker(copy(qubit_pauli_operator)),
                 )
                 simulator_wires.append(wire_sim)
                 device_wires.append(wire_device)
         return (experiment_wires, simulator_wires, device_wires)
 
     return MitTask(
-        _label="CCL_State_Circuits",
+        _label="CCLStateCircuits",
         _n_in_wires=1,
         _n_out_wires=3,
         _method=task,
@@ -312,7 +396,7 @@ def ccl_result_batching_task_gen(n_state_circuits: int) -> MitTask:
         experiment characterisation.
     :type n_state_circuits: int
 
-    :return: MitTask obect that organises QubitPauliOperator objects required for
+    :return: MitTask object that organises QubitPauliOperator objects required for
         characterisation.
     :rtype: MitTask.
     """
@@ -351,7 +435,7 @@ def ccl_likelihood_filtering_task_gen(
 ) -> MitTask:
     """
     :param likelihood_function: LikelihoodFunction enum used to accept or reject some pair of noisy and noiseless expectation.
-        Function must take two QubitPauliOperator as parametrer, and return a single float between 0 and 1 as answer.
+        Function must take two QubitPauliOperator as parameter, and return a single float between 0 and 1 as answer.
     :type likelihood_function: LikelihoodFunction
     :key seed: Seed value for sampling probability for likelihood function
 
@@ -433,34 +517,43 @@ def gen_CDR_MitEx(
     :key model: Model characterised by state circuits, default _PolyCDRCorrect(1) (see cdr_post.py for other options).
     :key likelihood_function: LikelihoodFunction used to filter state circuit results, given by a LikelihoodFunction Enum,
         default set to none.
+    :key tolerance: Model can be perturbed when calibration circuits have by
+        exact expectation values too close to each other. This parameter
+        sets a distance between exact expectation values which at least some
+        calibration circuits should have.
+    :key distance_tolerance: The absolute tolerance on the distance between
+        expectation values of the calibration and original circuit.
+    :key calibration_fraction: The upper bound on the fraction of calibration
+        circuits which have noisy expectation values far from that of the
+        original circuit.
     """
-    _states_sim_mitex = copy.copy(
+    _states_sim_mitex = copy(
         kwargs.get(
             "states_simluator_mitex",
             MitEx(
                 simulator_backend,
-                _label="StatesSimMitex",
+                _label="StatesSimMitEx",
                 mitres=gen_compiled_MitRes(simulator_backend, 0),
             ),
         )
     )
-    _states_device_mitex = copy.copy(
+    _states_device_mitex = copy(
         kwargs.get(
             "states_device_mitex",
             MitEx(
                 device_backend,
-                _label="StatesDeviceMitex",
+                _label="StatesDeviceMitEx",
                 mitres=gen_compiled_MitRes(device_backend, 0),
             ),
         )
     )
-    _experiment_mitex = copy.copy(
+    _experiment_mitex = copy(
         kwargs.get(
             "experiment_mitex",
             MitEx(
                 device_backend,
-                _label="ExperimentMitex",
-                mitres=gen_compiled_MitRes(device_backend, 2),
+                _label="ExperimentMitEx",
+                mitres=gen_compiled_MitRes(device_backend, 0),
             ),
         )
     )
@@ -470,18 +563,41 @@ def gen_CDR_MitEx(
     _states_sim_taskgraph.append(ccl_result_batching_task_gen(total_state_circuits))
 
     likelihood_function = kwargs.get("likelihood_function", LikelihoodFunction.none)
-    _states_sim_taskgraph.append(ccl_likelihood_filtering_task_gen(likelihood_function))
-    _states_sim_taskgraph.append(
-        cdr_calibration_task_gen(
-            device_backend, kwargs.get("model", _PolyCDRCorrect(1))
-        )
-    )
 
     _experiment_taskgraph = TaskGraph().from_TaskGraph(_experiment_mitex)
     _experiment_taskgraph.parallel(_states_sim_taskgraph)
-    _experiment_taskgraph.prepend(
-        ccl_state_task_gen(n_non_cliffords, n_pairs, total_state_circuits)
+
+    _post_calibrate_task_graph = TaskGraph(_label="FitCalibrate")
+    _post_calibrate_task_graph.append(
+        ccl_likelihood_filtering_task_gen(likelihood_function)
     )
+    _post_calibrate_task_graph.append(
+        cdr_calibration_task_gen(
+            device_backend,
+            kwargs.get("model", _PolyCDRCorrect(1)),
+        )
+    )
+
+    _post_task_graph = TaskGraph(_label="QualityCheckCorrect")
+    _post_task_graph.parallel(_post_calibrate_task_graph)
+    _post_task_graph.prepend(
+        cdr_quality_check_task_gen(
+            distance_tolerance=kwargs.get("distance_tolerance", 0.1),
+            calibration_fraction=kwargs.get("calibration_fraction", 0.5),
+        )
+    )
+
+    _experiment_taskgraph.prepend(
+        ccl_state_task_gen(
+            n_non_cliffords,
+            n_pairs,
+            total_state_circuits,
+            simulator_backend=simulator_backend,
+            tolerance=kwargs.get("tolerance", 0.01),
+            max_state_circuits_attempts=kwargs.get("max_state_circuits_attempts", 10),
+        )
+    )
+    _experiment_taskgraph.append(_post_task_graph)
     _experiment_taskgraph.append(cdr_correction_task_gen(device_backend))
 
     return MitEx(device_backend).from_TaskGraph(_experiment_taskgraph)
