@@ -12,9 +12,9 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-
 from qermit import (
     MitEx,
+    MitRes,
     AnsatzCircuit,
     MitTask,
     ObservableTracker,
@@ -24,24 +24,27 @@ from qermit import (
 from qermit.zero_noise_extrapolation.zne import (
     gen_duplication_task,
     gen_initial_compilation_task,
+    gen_qubit_relabel_task,
 )
 from qermit.probabilistic_error_cancellation.cliff_circuit_gen import (
     random_clifford_circ,
 )
 
-from pytket.passes import RebaseIBM, DecomposeBoxes  # type: ignore
+from sympy.core.expr import Expr
+
+from pytket.passes import RebaseTket, DecomposeBoxes  # type: ignore
 from pytket.utils import QubitPauliOperator, get_pauli_expectation_value
 from pytket.backends import Backend
 from pytket.transform import Transform  # type: ignore
 from pytket.circuit import Op, CircBox, OpType, Circuit, Node  # type: ignore
-from pytket.routing import place_with_map  # type: ignore
+from pytket.placement import place_with_map  # type: ignore
 
 from pytket.pauli import QubitPauliString  # type: ignore
 from pytket.predicates import CliffordCircuitPredicate  # type: ignore
 
 import re
 from typing import List, Tuple, Dict, cast, Union, Any
-import copy
+from copy import copy
 import numpy as np
 
 QuasiProbabilities = List[float]
@@ -71,6 +74,7 @@ def random_commuting_clifford(
     qps: QubitPauliString,
     simulator_backend: Backend,
     max_count: int = 1000,
+    n_shots: int = 1000,
 ) -> Circuit:
     """Replace all Computing gates with random Clifford gates. The expectation
     of the given Pauli string on the final Clifford circuit is non-zero.
@@ -110,16 +114,13 @@ def random_commuting_clifford(
     # of iterations is exceeded.
     expect_val = complex(0)
     while round(abs(expect_val)) == 0:
-
         rand_cliff_circ = circ.copy()
 
         # Retrieve a list of random Clifford gates, one for each of
         # the Computing gates in the original circuit. Note this is in the form of a
         # CircBox so that the opgroup labels persist after substitution
         # (they would not do so if circuits were used instead of CircBox)
-        rand_cliff_list = [
-            CircBox(random_clifford_circ(1)) for opgroup in comp_opgroup_list
-        ]
+        rand_cliff_list = [CircBox(random_clifford_circ(1)) for _ in comp_opgroup_list]
         # Replace Computing gates with Clifford gates.
         for opgroup, rand_cliff in zip(comp_opgroup_list, rand_cliff_list):
             rand_cliff_circ.substitute_named(rand_cliff, opgroup)
@@ -136,19 +137,28 @@ def random_commuting_clifford(
 
         new_qps_qbs = []
         qps_paulis = []
-        qps_dict = qps.to_dict()
+        qps_dict = qps.map
         for x in qps_dict:
             new_qps_qbs.append(n_q_map[x])
             qps_paulis.append(qps_dict[x])
 
         new_qps = QubitPauliString(new_qps_qbs, qps_paulis)
 
-        rand_cliff_circ_copy = rand_cliff_circ.copy()
-        place_with_map(rand_cliff_circ_copy, n_q_map)
+        place_with_map(rand_cliff_circ, n_q_map)
 
-        expect_val = get_pauli_expectation_value(
-            rand_cliff_circ_copy, new_qps, simulator_backend
-        )
+        # Check if state is supported, otherwise use shots, otherwise raise error
+        if simulator_backend.supports_state:
+            expect_val = get_pauli_expectation_value(
+                rand_cliff_circ, new_qps, simulator_backend
+            )
+        elif simulator_backend.supports_shots or simulator_backend.supports_counts:
+            expect_val = get_pauli_expectation_value(
+                rand_cliff_circ, new_qps, simulator_backend, n_shots=n_shots
+            )
+        else:
+            raise RuntimeError(
+                "The simulator backend does not support state, shots or counts."
+            )
         # TODO: Better management of the case that there are no circuits with expectation value not equal to 0.
 
         # Check if the number of attempts at finding a circuit with non-zero expectation exceeds the maximum.
@@ -253,7 +263,7 @@ def PECRebase(circ: Circuit) -> Circuit:
     :rtype: Circuit
     """
     rebased_circ = circ.copy()
-    RebaseIBM().apply(rebased_circ)
+    RebaseTket().apply(rebased_circ)
     Transform.ReduceSingles().apply(rebased_circ)
     return rebased_circ
 
@@ -362,7 +372,7 @@ def gen_run_with_quasi_prob() -> MitTask:
                 for j, prob in enumerate(circuit_prob[i]):
                     em_expect[qps] += prob * circuit_results[j][qps]
             em_expect_qpo = QubitPauliOperator(
-                cast(Dict[Any, Union[int, float, complex]], em_expect)
+                cast(Dict[Any, Union[int, float, complex, Expr]], em_expect)
             )
             em_expect_list.append(em_expect_qpo)
 
@@ -376,14 +386,12 @@ def gen_run_with_quasi_prob() -> MitTask:
     )
 
 
-def collate_results_task_gen(num_cliff_circ: int) -> MitTask:
+def collate_results_task_gen() -> MitTask:
     """Generates task which collates results from running circuit, and circuits with frame
     gates wrapped in Pauli gates. The results are collated so as to facilitate
     learning the quasiprobabilities required for correction. The data itself is
     not changed by this task.
 
-    :param num_cliff_circ: The number of Clifford circuits generated for each inputted circuit.
-    :type num_cliff_circ: int
     :return: MitTask object collating results.
     :rtype: MitTask
     """
@@ -686,7 +694,7 @@ def label_gates(circ: Circuit) -> Circuit:
     """Label all of the gates in the circuit as with "Frame" or "Computing".
     The label includes an index to describe the ordering of the gates
 
-    :param circ: Circuit which should be in the U1, U2, U3, CX basis
+    :param circ: Circuit which should be in the TK1, CX basis
     :type circ: Circuit
     :raises RuntimeError: Raised if the circuit is not in the required basis.
     :return: Identical circuit, but with gates assigned opgroups.
@@ -703,7 +711,7 @@ def label_gates(circ: Circuit) -> Circuit:
     frame_count = 0
     for command in command_list:
         labelled_command = command.copy()
-        if labelled_command["op"]["type"] in ("U1", "U2", "U3"):
+        if labelled_command["op"]["type"] in ("TK1"):
             labelled_command["opgroup"] = "Computing %i" % comp_count
             comp_count += 1
         elif labelled_command["op"]["type"] in ("CX"):
@@ -711,7 +719,7 @@ def label_gates(circ: Circuit) -> Circuit:
             frame_count += 1
         else:
             raise RuntimeError(
-                'This gate is not one of either "U1", "U2", "U3" or "CX". Please ensure you have run PECRebase before using this function.'
+                'This gate is not one of either "TK1" or "CX". Please ensure you have run PECRebase before using this function.'
             )
         labelled_command_list.append(labelled_command)
 
@@ -975,13 +983,14 @@ def gen_get_noisy_circuits(backend: Backend, **kwargs) -> MitTask:
                     [str_to_pauli_op(error["op"][0]), str_to_pauli_op(error["op"][1])],
                 )
 
-                backend.compile_circuit(pauli_circ, optimisation_level=0)
+                pauli_circ = backend.get_compiled_circuit(
+                    pauli_circ, optimisation_level=0
+                )
 
-                # TODO: Replace with a copy of MeasurementSetup rather than new object
                 new_ansatz_circuit = AnsatzCircuit(
                     Circuit=pauli_circ,
-                    Shots=copy.copy(experiment.AnsatzCircuit.Shots),
-                    SymbolsDict=copy.copy(experiment.AnsatzCircuit.SymbolsDict),
+                    Shots=copy(experiment.AnsatzCircuit.Shots),
+                    SymbolsDict=copy(experiment.AnsatzCircuit.SymbolsDict),
                 )
                 new_tracker = ObservableTracker(
                     experiment.ObservableTracker.qubit_pauli_operator
@@ -1044,14 +1053,18 @@ def gen_PEC_learning_based_MitEx(
     # TODO: Change to a number of clifford circuits which varies with the size of the circuit
     num_cliff_circ = kwargs.get("num_cliff", 10)
 
-    sim_mitex = copy.copy(
+    sim_mitex = copy(
         kwargs.get(
             "simulator_mitex", MitEx(simulator_backend, _label="IdealCliffordMitEx")
         )
     )
 
-    device_mitex = copy.copy(
-        kwargs.get("device_mitex", MitEx(device_backend, _label="NoisyMitex"))
+    device_mitres = MitRes(device_backend)
+    device_mitex = copy(
+        kwargs.get(
+            "device_mitex",
+            MitEx(device_backend, _label="NoisyMitex", mitres=device_mitres),
+        )
     )
 
     _experiment_taskgraph = TaskGraph().from_TaskGraph(device_mitex)
@@ -1074,7 +1087,7 @@ def gen_PEC_learning_based_MitEx(
     )
     _experiment_taskgraph.prepend(get_clifford_training_set)
 
-    collate_results = collate_results_task_gen(num_cliff_circ)
+    collate_results = collate_results_task_gen()
     _experiment_taskgraph.append(collate_results)
 
     learn_dist = learn_quasi_probs_task_gen(num_cliff_circ)
@@ -1104,6 +1117,9 @@ def gen_PEC_learning_based_MitEx(
     _experiment_taskgraph.prepend(wrap_frame_gates)
     _experiment_taskgraph.prepend(label_gates)
     _experiment_taskgraph.prepend(compile_to_frames_and_computing)
+
+    _experiment_taskgraph.add_wire()
     _experiment_taskgraph.prepend(initial_compilation)
+    _experiment_taskgraph.append(gen_qubit_relabel_task())
 
     return MitEx(device_backend).from_TaskGraph(_experiment_taskgraph)
