@@ -11,8 +11,6 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
-
-
 from pytket.backends import Backend
 from qermit import (
     MitEx,
@@ -23,17 +21,28 @@ from qermit import (
     ObservableExperiment,
     TaskGraph,
 )
-import copy
+from copy import copy
 from pytket.pauli import QubitPauliString  # type: ignore
 from enum import Enum
 import numpy as np
 from scipy.optimize import curve_fit  # type: ignore
-from typing import List, Tuple, cast
-from pytket import Circuit
+from typing import List, Tuple, cast, Dict
+from pytket import Circuit, OpType
 from pytket.predicates import CompilationUnit  # type: ignore
 from pytket.utils import QubitPauliOperator
 import matplotlib.pyplot as plt  # type: ignore
 from numpy.polynomial.polynomial import Polynomial
+from pytket.circuit import Node  # type: ignore
+
+
+box_types = {
+    OpType.CircBox,
+    OpType.ExpBox,
+    OpType.PauliExpBox,
+    OpType.Unitary1qBox,
+    OpType.Unitary2qBox,
+    OpType.Unitary3qBox,
+}
 
 
 class Folding(Enum):
@@ -44,10 +53,9 @@ class Folding(Enum):
     :rtype: Circuit
     """
 
-    # TODO: Understand why self is needed etc
     # TODO: circ does not appear as input in docs
     # TODO Generalise with 'partial folding' to allow for non integer noise scaling
-    def circuit(circ: Circuit, noise_scaling: int, *args) -> Circuit:
+    def circuit(circ: Circuit, noise_scaling: int, **kwargs) -> Circuit:
         """Noise scaling by circuit folding. In this case the folded circuit is of
         the form :math:`CC^{-1}CC^{-1}...C` where :math:`C` is the original circuit. As such noise may be scaled by
         odd integers. The Unitary implemented is unchanged by this process.
@@ -72,14 +80,31 @@ class Folding(Enum):
 
             # Add barrier between circuit and its inverse
             folded_circ.add_barrier(folded_circ.qubits + folded_circ.bits)
-            folded_circ.add_circuit(circ.dagger(), folded_circ.qubits, folded_circ.bits)
 
+            # Add inverse circuit by iterating though commands and inverting them
+            for gate in reversed(circ.get_commands()):
+                if gate.op.type == OpType.Barrier:
+                    folded_circ.add_barrier(gate.args)
+                elif gate.op.type in box_types:
+                    raise RuntimeError("Box types not supported when folding.")
+                else:
+                    folded_circ.add_gate(gate.op.dagger, gate.args)
+
+            # Add barrier between circuit and its inverse
             folded_circ.add_barrier(folded_circ.qubits + folded_circ.bits)
-            folded_circ.add_circuit(circ, folded_circ.qubits, folded_circ.bits)
+
+            # Add original circuit
+            for gate in circ.get_commands():
+                if gate.op.type == OpType.Barrier:
+                    folded_circ.add_barrier(gate.args)
+                elif gate.op.type in box_types:
+                    raise RuntimeError("Box types not supported when folding.")
+                else:
+                    folded_circ.add_gate(gate.op, gate.args)
 
         return folded_circ
 
-    def gate(circ: Circuit, noise_scaling: float, _allow_approx_fold: bool) -> Circuit:
+    def gate(circ: Circuit, noise_scaling: float, **kwargs) -> Circuit:
         """Noise scaling by gate folding. In this case gates :math:`G` are replaced at random
         with :math:`GG^{-1}G` until the number of gates is sufficiently scaled.
 
@@ -87,26 +112,39 @@ class Folding(Enum):
         :type circ: Circuit
         :param noise_scaling: Factor by which to increase the noise.
         :type noise_scaling: float
-        :param _allow_approx_fold: Allows for the noise to be increased by an amount close to that requested, as
+
+        :key _allow_approx_fold: Allows for the noise to be increased by an amount close to that requested, as
             opposed to by exactly the amount requested.
             This is necessary as there are cases where the exact noise scaling cannot be achieved.
             This occurs due to the discrete
             amounts by which the noise can be increased (i.e. the discrete amount by which one gate increases the noise).
         :type _allow_approx_fold: bool
+
         :raises ValueError: Raised if the requested noise scaling cannot be exactly achieved. This can be
             avoided by appropriately setting _allow_approx_fold.
         :return: Folded circuit implementing identical unitary to the initial circuit.
         :rtype: Circuit
         """
 
+        _allow_approx_fold = kwargs.get("_allow_approx_fold", 0)
+
         c_dict = circ.to_dict()
         num_commands = len(c_dict["commands"])
 
-        # Calculate the number of gates that need to be folded
-        num_additional_commands = (noise_scaling - 1) * num_commands
+        # Count and locate the number of commands that are not barriers
+        c_dict_commands_non_barrier = [
+            (i, command)
+            for (i, command) in enumerate(c_dict["commands"])
+            if command["op"]["type"] != "Barrier"
+        ]
+        num_non_barrier_commands = len(c_dict_commands_non_barrier)
+
+        # Calculate the number of gates that need to be folded. Note that only
+        # non barrier commands should be folded so only these are counted.
+        num_additional_commands = (noise_scaling - 1) * num_non_barrier_commands
         num_folded_commands = int(num_additional_commands // 2)
 
-        true_noise_scaling = 1 + ((num_folded_commands * 2) / num_commands)
+        true_noise_scaling = 1 + ((num_folded_commands * 2) / num_non_barrier_commands)
         # This check isolates the case where the noise folding that can be achieved is different
         # from that requested. While noise_scaling can be any real value, as the gates increase
         # the noise by discrete values, not all folding values are possible.
@@ -118,32 +156,57 @@ class Folding(Enum):
             )
 
         # Choose a random selection of commands to fold. Commands are referenced by their index.
-        # These are chosen with replacement.
+        # These are chosen with replacement. Only non barrier commands should
+        # be folded, and so only the index of those are added here.
         commands_to_fold = np.random.choice(
-            [i for i in range(num_commands)], num_folded_commands
+            [i[0] for i in c_dict_commands_non_barrier], num_folded_commands
         )
         # Calculate how many times each individual command needs to be folded
         num_folds = {i: list(commands_to_fold).count(i) for i in range(num_commands)}
 
+        command_circ_dict = {
+            key: val for key, val in c_dict.items() if key != "commands"
+        }
+
         folded_command_list = []
         # For each command, fold the appropriate number of times.
         for command_index in num_folds:
 
             command = c_dict["commands"][command_index]
-
-            command_circ_dict = c_dict.copy()
-            command_circ_dict["commands"] = [command]
+            command_circ_dict.update({"commands": [command]})
             command_circ = Circuit().from_dict(command_circ_dict)
 
-            # Find the inverse of the command
-            inverse_command_circ = command_circ.dagger()
-            inverse_command_circ_dict = inverse_command_circ.to_dict()
-            inverse_command = inverse_command_circ_dict["commands"]
+            # Commands which are not to be folded may not be invertible (for
+            # example barriers) and so the inverse is not calculated in that
+            # case.
+            if num_folds[command_index] > 0:
+                # Find the inverse of the command
+                inverse_command_circ = command_circ.dagger()
+                inverse_command_circ_dict = inverse_command_circ.to_dict()
+                inverse_command = inverse_command_circ_dict["commands"]
 
             # Append command and inverse the appropriate number of times.
             folded_command_list.append(command)
             for _ in range(num_folds[command_index]):
+                folded_command_list.append(
+                    {
+                        "args": command["args"],
+                        "op": {
+                            "signature": ["Q"] * len(command["args"]),
+                            "type": "Barrier",
+                        },
+                    }
+                )
                 folded_command_list.append(*inverse_command)
+                folded_command_list.append(
+                    {
+                        "args": command["args"],
+                        "op": {
+                            "signature": ["Q"] * len(command["args"]),
+                            "type": "Barrier",
+                        },
+                    }
+                )
                 folded_command_list.append(command)
 
         folded_c_dict = c_dict.copy()
@@ -152,6 +215,80 @@ class Folding(Enum):
 
         return folded_c
 
+    def odd_gate(circ: Circuit, noise_scaling: int, **kwargs) -> Circuit:
+        """Noise scaling by gate folding. In this case odd gates :math:`G` are
+        replaced :math:`GG^{-1}G` until the number of gates is sufficiently
+        scaled.
+
+        :param circ: Original circuit to be folded.
+        :type circ: Circuit
+        :param noise_scaling: Factor by which to increase the noise.
+        :type noise_scaling: float
+        :return: Folded circuit implementing identical unitary to the initial circuit.
+        :rtype: Circuit
+        """
+
+        c_dict = circ.to_dict()
+
+        fold = True
+        folded_command_list = []
+
+        command_circ_dict = {
+            key: val for key, val in c_dict.items() if key != "commands"
+        }
+
+        for command in c_dict["commands"]:
+
+            # Barriers are added to the circuit but otherwise effectively
+            # skipped.
+            if command["op"]["type"] == "Barrier":
+
+                folded_command_list.append(command)
+
+            elif fold:
+                command_circ_dict.update({"commands": [command]})
+                command_circ = Circuit.from_dict(command_circ_dict)
+
+                # Find the inverse of the command
+                inverse_command_circ = command_circ.dagger()
+                inverse_command_circ_dict = inverse_command_circ.to_dict()
+                inverse_command = inverse_command_circ_dict["commands"]
+
+                # Append command and inverse the appropriate number of times.
+                folded_command_list.append(command)
+                for _ in range(noise_scaling - 1):
+                    folded_command_list.append(
+                        {
+                            "args": command["args"],
+                            "op": {
+                                "signature": ["Q"] * len(command["args"]),
+                                "type": "Barrier",
+                            },
+                        }
+                    )
+                    folded_command_list.append(*inverse_command)
+                    folded_command_list.append(
+                        {
+                            "args": command["args"],
+                            "op": {
+                                "signature": ["Q"] * len(command["args"]),
+                                "type": "Barrier",
+                            },
+                        }
+                    )
+                    folded_command_list.append(command)
+
+                fold = not fold
+            else:
+                folded_command_list.append(command)
+
+                fold = not fold
+
+        c_dict.update({"commands": folded_command_list})
+        folded_c = Circuit.from_dict(c_dict)
+
+        return folded_c
+
 
 def poly_exp_func(x: float, *params) -> float:
     """Definition of poly-exponential function for the purposes of fitting to data
@@ -210,7 +347,7 @@ class Fit(Enum):
         # Plot fitted function and data
         if _show_fit:
 
-            fit_x = np.linspace(0, x[-1], 100)
+            fit_x = np.linspace(0, max(x), 100)
             fit_y = [cube_root_func(i, *vals[0]) for i in fit_x]
 
             plot_fit(x, y, cast(List[float], fit_x), fit_y, fit_to_zero)
@@ -247,8 +384,24 @@ class Fit(Enum):
             )
 
         # Fit data to polyexponential function
+        # TODO: Improve bounds here.
+        bounds = (
+            [-1, -2, *[-np.inf for i in range(deg)]],
+            [1, 2, *[np.inf for i in range(deg)]],
+        )
+
+        # Initialise decaying poly-exponential with intersect at
+        # unfolded noisy value.
+        least_noisy_y_index = x.index(1)
+        p0 = [0, y[least_noisy_y_index], *[-1 for i in range(deg)]]
+
         vals = curve_fit(
-            poly_exp_func, x, y, p0=[0, 1, *[-1 for i in range(deg)]], maxfev=10000
+            poly_exp_func,
+            x,
+            y,
+            p0=p0,
+            maxfev=10000,
+            bounds=bounds,
         )
 
         # Extrapolate function to zero noise limit
@@ -257,7 +410,7 @@ class Fit(Enum):
         # Plot data and fitted function
         if _show_fit:
 
-            fit_x = np.linspace(0, x[-1], 100)
+            fit_x = np.linspace(0, max(x), 100)
             fit_y = [poly_exp_func(i, *vals[0]) for i in fit_x]
 
             plot_fit(x, y, cast(List[float], fit_x), fit_y, fit_to_zero)
@@ -443,14 +596,14 @@ def digital_folding_task_gen(
         for experiment in mitex_wire:
 
             # Apply the necessary folding method
-            zne_circ = _folding_type(experiment.AnsatzCircuit.Circuit, noise_scaling, _allow_approx_fold)  # type: ignore
+            zne_circ = _folding_type(experiment.AnsatzCircuit.Circuit, noise_scaling, _allow_approx_fold=_allow_approx_fold)  # type: ignore
 
             # TODO: This additional compilation pass may result in the circuit noise being
             # increased too much, and should be removed or better accounted for.
 
             # This compilation pass was added to account for the case that
             # the inverse of a gate is not in the gateset of the backend.
-            backend.default_compilation_pass(optimisation_level=0).apply(zne_circ)
+            backend.rebase_pass().apply(zne_circ)
 
             folded_circuits.append(
                 ObservableExperiment(
@@ -537,6 +690,32 @@ def extrapolation_task_gen(
     )
 
 
+def copy_mitex_wire(wire: ObservableExperiment) -> ObservableExperiment:
+    """Returns a single copy of the inputted wire
+
+    :param wire: Pair of ansatz circuit and ObservableTracker
+    :type wire: ObservableExperiment
+    :return: single copy of inputted wire
+    :rtype: ObservableExperiment
+    """
+
+    # Copy ansatz circuit
+    new_ansatz_circuit = AnsatzCircuit(
+        Circuit=wire.AnsatzCircuit.Circuit.copy(),
+        Shots=copy(wire.AnsatzCircuit.Shots),
+        SymbolsDict=copy(wire.AnsatzCircuit.SymbolsDict),
+    )
+
+    # copy qps and instantiate new measurement setup
+    new_ot = ObservableTracker(
+        QubitPauliOperator(copy(wire.ObservableTracker._qubit_pauli_operator._dict))
+    )
+    new_wire = ObservableExperiment(
+        AnsatzCircuit=new_ansatz_circuit, ObservableTracker=new_ot
+    )
+    return new_wire
+
+
 def gen_duplication_task(duplicates: int, **kwargs) -> MitTask:
     """Duplicate the inputted experiment wire
 
@@ -563,33 +742,6 @@ def gen_duplication_task(duplicates: int, **kwargs) -> MitTask:
                 "The number of duplications must be greater than or equal to 1."
             )
 
-        def copy_mitex_wire(wire: ObservableExperiment) -> ObservableExperiment:
-            """Returns a single copy of the inputted wire
-
-            :param wire: Pair of ansatz circuit and ObservableTracker
-            :type wire: ObservableExperiment
-            :return: single copy of inputted wire
-            :rtype: ObservableExperiment
-            """
-
-            # Copy ansatz circuit
-            new_ansatz_circuit = AnsatzCircuit(
-                Circuit=wire.AnsatzCircuit.Circuit.copy(),
-                Shots=copy.copy(wire.AnsatzCircuit.Shots),
-                SymbolsDict=copy.copy(wire.AnsatzCircuit.SymbolsDict),
-            )
-
-            # copy qps and instantiate new measurement setup
-            new_ot = ObservableTracker(
-                QubitPauliOperator(
-                    copy.copy(wire.ObservableTracker._qubit_pauli_operator._dict)
-                )
-            )
-            new_wire = ObservableExperiment(
-                AnsatzCircuit=new_ansatz_circuit, ObservableTracker=new_ot
-            )
-            return new_wire
-
         # Compose coppies into wire format
         if duplicates == 1:
             me_copy = [copy_mitex_wire(circuit_tuple) for circuit_tuple in mitex_wire]
@@ -611,10 +763,36 @@ def gen_duplication_task(duplicates: int, **kwargs) -> MitTask:
     )
 
 
+def qpo_node_relabel(qpo: QubitPauliOperator, node_map: Dict[Node, Node]):
+    """Relabel the nodes of qpo according to node_map
+
+    :param qpo: Original qubit pauli operator
+    :type qpo: QubitPauliOperator
+    :param node_map: Map between nodes
+    :type node_map: Dict[Node,Node]
+    :return: Relabeled qubit pauli operator
+    :rtype: QubitPauliOperator
+    """
+
+    orig_qpo_dict = qpo._dict.copy()
+    new_qpo_dict = {}
+    for orig_qps in orig_qpo_dict:
+        orig_qps_dict = orig_qps.map
+        new_qps_dict = {}
+        for q in orig_qps_dict:
+            new_qps_dict[node_map[q]] = orig_qps_dict[q]
+        new_qps = QubitPauliString(new_qps_dict)
+        new_qpo_dict[new_qps] = orig_qpo_dict[orig_qps]
+
+    return QubitPauliOperator(new_qpo_dict)
+
+
 def gen_initial_compilation_task(
     backend: Backend, optimisation_level: int = 1
 ) -> MitTask:
-    """Perform compilation before folding.
+    """Perform compilation to the backend. Note that this will relabel the
+    nodes of the device, and so should be followed by gen_qubit_relabel_task
+    in the task graph.
 
     :param backend: Backend to compile to
     :type backend: Backend
@@ -624,14 +802,16 @@ def gen_initial_compilation_task(
 
     def task(
         obj, wire: List[ObservableExperiment]
-    ) -> Tuple[List[ObservableExperiment]]:
+    ) -> Tuple[List[ObservableExperiment], Dict[Node, Node]]:
         """Performs initial compilation before folding. This is to ensure minimal compilation
         after folding, as this could disrupt by how much the noise is increased.
 
         :param wire: List of experiments
         :type wire: List[ObservableExperiment]
-        :return: List of experiments compiled to run on the inputted backend
-        :rtype: Tuple[List[ObservableExperiment]]
+        :return: List of experiments compiled to run on the inputted backend.
+        Additionally a dictionary describing how the nodes have been mapped
+        by compilation.
+        :rtype: Tuple[List[ObservableExperiment], Dict[Node, Node]]
         """
 
         mapped_wire = []
@@ -641,27 +821,16 @@ def gen_initial_compilation_task(
             # Perform default compilation, tracking to which physical
             # qubits the initial qubits are mapped
             compiled_circ = obs_exp.AnsatzCircuit.Circuit.copy()
+
             cu = CompilationUnit(compiled_circ)
             backend.default_compilation_pass(
                 optimisation_level=optimisation_level
             ).apply(cu)
-            final_map = cu.final_map
-
-            # TODO: This map needs to be used at the end of the
-            # process in order to map the output pauli string back
-            # to the pauli strings inputted by the user
+            node_map = cu.final_map
 
             # Alter the qubit pauli operator so that it maps to the new physical qubits.
-            orig_qpo_dict = obs_exp[1]._qubit_pauli_operator._dict.copy()
-            new_qpo_dict = {}
-            for orig_qps in orig_qpo_dict:
-                orig_qps_dict = orig_qps.to_dict()
-                new_qps_dict = {}
-                for q in orig_qps_dict:
-                    new_qps_dict[final_map[q]] = orig_qps_dict[q]
-                new_qps = QubitPauliString(new_qps_dict)
-                new_qpo_dict[new_qps] = orig_qpo_dict[orig_qps]
-            new_qpo = QubitPauliOperator(new_qpo_dict)
+            qpo = obs_exp[1]._qubit_pauli_operator
+            new_qpo = qpo_node_relabel(qpo, node_map)
 
             # Construct new list of experiments, but with compiled circuits.
             mapped_wire.append(
@@ -675,16 +844,55 @@ def gen_initial_compilation_task(
                 )
             )
 
-        return (mapped_wire,)
+        return (
+            mapped_wire,
+            node_map,
+        )
 
     return MitTask(
         _label="CompileToBackend",
-        _n_out_wires=1,
+        _n_out_wires=2,
         _n_in_wires=1,
         _method=task,
     )
 
 
+def gen_qubit_relabel_task() -> MitTask:
+    """Task reversing the relabelling of qubits performed during compilation.
+    This should follow gen_initial_compilation_task
+
+    :return: Task performing relabelling.
+    :rtype: MitTask
+    """
+
+    def task(
+        obj, qpo_list: List[QubitPauliOperator], compilation_map: Dict[Node, Node]
+    ) -> Tuple[List[QubitPauliOperator]]:
+        """Use node map returned by compilation unit to undo the relabelling
+        performed by gen_initial_compilation_task
+
+        :param qpo_list: List of QubitPauliOperator
+        :type qpo_list: List[QubitPauliOperator]
+        :param compilation_map: Dictionary mapping nodes as returned by
+        gen_initial_compilation_task task
+        :type compilation_map: Dict[Node, Node]
+        :return: List of QubitPauliOperator with relabeled nodes.
+        :rtype: Tuple[List[QubitPauliOperator]]
+        """
+
+        node_map = {value: key for key, value in compilation_map.items()}
+        new_qpo_list = [qpo_node_relabel(qpo, node_map) for qpo in qpo_list]
+
+        return (new_qpo_list,)
+
+    return MitTask(
+        _label="RelabelQubits",
+        _n_out_wires=1,
+        _n_in_wires=2,
+        _method=task,
+    )
+
+
 # TODO: Backend does not appear as input in documentation
 def gen_ZNE_MitEx(backend: Backend, noise_scaling_list: List[float], **kwargs) -> MitEx:
     """Generates MitEx object which mitigates for noise using Zero Noise Extrapolation. This is the
@@ -698,14 +906,14 @@ def gen_ZNE_MitEx(backend: Backend, noise_scaling_list: List[float], **kwargs) -
     :return: MitEx object performing noise mitigation by ZNE.
     :rtype: MitEx
     """
-    _experiment_mitres = copy.copy(
+    _experiment_mitres = copy(
         kwargs.get(
             "experiment_mitres",
             MitRes(backend),
         )
     )
 
-    _experiment_mitex = copy.copy(
+    _experiment_mitex = copy(
         kwargs.get(
             "experiment_mitex",
             MitEx(backend, _label="ExperimentMitex", mitres=_experiment_mitres),
@@ -715,8 +923,8 @@ def gen_ZNE_MitEx(backend: Backend, noise_scaling_list: List[float], **kwargs) -
 
     _optimisation_level = kwargs.get("optimisation_level", 0)
     _show_fit = kwargs.get("show_fit", False)
-    _folding_type = copy.copy(kwargs.get("folding_type", Folding.circuit))
-    _fit_type = copy.copy(kwargs.get("fit_type", Fit.linear))
+    _folding_type = kwargs.get("folding_type", Folding.circuit)
+    _fit_type = kwargs.get("fit_type", Fit.linear)
     _deg = kwargs.get("deg", len(noise_scaling_list) - 1)
     _seed = kwargs.get("seed", None)
     _allow_approx_fold = kwargs.get("allow_approx_fold", True)
@@ -727,25 +935,37 @@ def gen_ZNE_MitEx(backend: Backend, noise_scaling_list: List[float], **kwargs) -
 
         _label = str(fold) + "FoldMitEx"
 
-        fold_mitex = copy.copy(
+        _fold_mitres = copy(
             kwargs.get(
-                "fold_mitex", MitEx(backend, _label=_label, mitres=MitRes(backend))
+                "experiment_mitres",
+                MitRes(backend),
+            )
+        )
+
+        _fold_mitex = copy(
+            kwargs.get(
+                "experiment_mitex",
+                MitEx(backend, _label=_label, mitres=_fold_mitres),
             )
         )
 
         digital_folding_task = digital_folding_task_gen(
             backend, fold, _folding_type, _allow_approx_fold
         )
-        fold_mitex.prepend(digital_folding_task)
-        _experiment_taskgraph.parallel(fold_mitex)
+        _fold_mitex.prepend(digital_folding_task)
+        _experiment_taskgraph.parallel(_fold_mitex)
 
     extrapolation_task = extrapolation_task_gen(
         noise_scaling_list, _fit_type, _show_fit, _deg
     )
 
-    _experiment_taskgraph.append(extrapolation_task)
     _experiment_taskgraph.prepend(gen_duplication_task(len(noise_scaling_list) + 1))
+    _experiment_taskgraph.append(extrapolation_task)
+
+    _experiment_taskgraph.add_wire()
+
     _experiment_taskgraph.prepend(
         gen_initial_compilation_task(backend, _optimisation_level)
     )
+    _experiment_taskgraph.append(gen_qubit_relabel_task())
     return MitEx(backend).from_TaskGraph(_experiment_taskgraph)
